
-- constructs the memory layout of a compiled plan
create function CompPlanDetails(@current_plan_address int, @plan_handle varbinary(64)) 
returns @details table 
(
      plan_handle varbinary(64) null,
      memory_object_address varbinary(8) not null,
      parent_address varbinary(8) null,
      type nvarchar(60) null,
      name nvarchar(256) null,
      pages_allocated_count int not null,
      page_size_in_bytes int not null,
      page_allocator_address varbinary(8) not null
)
as
begin
      -- Get the plan handle
      if (@plan_handle is null)     
select @plan_handle = plan_handle from sys.dm_exec_cached_plans where memory_object_address = @current_plan_address;
 
      -- Get all top level pmos into a temp table.
      with TopLevelPMOs as
      (
select @plan_handle as plan_handle, mo2.memory_object_address, mo2.parent_address, mo2.type, mo2.name, mo2.pages_allocated_count, mo2.page_size_in_bytes, mo2.page_allocator_address
      from sys.dm_os_memory_objects mo join sys.dm_os_memory_objects
mo2
      on mo2.page_allocator_address = mo.page_allocator_address
      where mo.memory_object_address = @current_plan_address
      )
 
-- find sub-pmos from all top level pmos, and add them plus the top level pmos into the result table.
      insert @details
select @plan_handle, mo3.memory_object_address, mo3.parent_address, mo3.type, mo3.name, mo3.pages_allocated_count, mo3.page_size_in_bytes, mo3.page_allocator_address
      from sys.dm_os_memory_objects mo3 join TopLevelPMOs
      on mo3.parent_address = TopLevelPMOs.memory_object_address
      union all
      select * from TopLevelPMOs
      return
end
go

create proc PlanBreakdown(@plan_addr varbinary(8) = null) as
begin
      set nocount on
 
      -- results tables
      if object_id('PlanDetails', 'U') is not null
            drop table PlanDetails
 
      if object_id('PlanSummary', 'U') is not null
            drop table PlanSummary
 
      create table PlanDetails
      (
            plan_handle varbinary(64) null,
            memory_object_address varbinary(8) not null,
            parent_address varbinary(8) null,
            type nvarchar(60) null,
            name nvarchar(256) null,
            pages_allocated_count int not null,
            page_size_in_bytes int not null,
            page_allocator_address varbinary(8) not null
      )
 
      create table PlanSummary
      (
            plan_handle varbinary(64) null,
            memory_object_address varbinary(8) not null,
            total_size int null,
            cstmt_count int not null,
            cstmt_size int not null,
            xstmt_count int not null,
            xstmt_size int not null,
            cursor_count int not null,
            cursor_size int not null,
            mxc_count int not null,
            mxc_size int not null,
            query_count int not null,
            query_size int not null,
            skeleton_count int not null
      )
 
-- temp table to hold the address of every plan that we're going to dump info for.
      create table #plans(plan_address varbinary(8))
 
      if (@plan_addr is null)
insert into #plans select memory_object_address from sys.dm_os_memory_objects
where type = 'MEMOBJ_COMPILE_ADHOC' and parent_address is NULL
      else
            insert into #plans values (@plan_addr)
      
declare plan_cursor cursor local for select plan_address from #plans
      open plan_cursor
 
      declare @current_plan_address varbinary(8)
      fetch next from plan_cursor into @current_plan_address
 
      while (@@fetch_status = 0)
      begin
            declare @plan_handle varbinary(64)
            declare @total_size int
            declare @cstmt_count int
            declare @cstmt_size int
            declare @xstmt_count int
            declare @xstmt_size int
            declare @cursor_count int
            declare @cursor_size int
            declare @mxc_count int
            declare @mxc_size int
            declare @query_count int
            declare @query_size int
            declare @skeleton_count int
            set @plan_handle = null;
            set @total_size = null;
            set @cstmt_count = 0;
            set @cstmt_size = 0;
            set @xstmt_count = 0;
            set @xstmt_size = 0;
            set @cursor_count = 0;
            set @cursor_size = 0;
            set @mxc_count = 0;
            set @mxc_size = 0;
            set @query_count = 0;
            set @query_size = 0;
            set @skeleton_count = 0;
            
-- grab the plan handle and total size for the plan we're working on, if it exists
            -- in dm_exec_cached_plans.
select @plan_handle = plan_handle, @total_size = size_in_bytes from
sys.dm_exec_cached_plans where memory_object_address = @current_plan_address
            
select * into #plan_details from CompPlanDetails(@current_plan_address, @plan_handle)
      
            -- add the plan details to the results
            insert into PlanDetails select * from #plan_details
 
-- If we didn't find this plan in sys.dm_exec_cached_plans, then go ahead and compute
-- the total_size by aggregating the pages used of all the top-level pmo's in the plan.
-- This really should be an accurate accounting of the plan memory usage too.
            if (@plan_handle is null)
select @total_size = sum(pages_allocated_count * page_size_in_bytes) from #plan_details
                  where parent_address is null
 
            -- form the summary information for the given plan.
            declare detail_cursor cursor local for 
select type, pages_allocated_count, page_size_in_bytes from #plan_details
            open detail_cursor
 
            declare @type varchar(256)
            declare @pages int      
            declare @page_size int
            
fetch next from detail_cursor into @type, @pages, @page_size
 
            while (@@fetch_status = 0)
            begin
                  if (@type = 'MEMOBJ_XSTMT')
                  begin
                        set @xstmt_count = @xstmt_count + 1
set @xstmt_size = @xstmt_size + (@pages * @page_size)
                  end
                  else if (@type = 'MEMOBJ_EXECUTE')
                  begin
                        set @mxc_count = @mxc_count + 1
set @mxc_size = @mxc_size + (@pages * @page_size)
                  end
                  else if (@type = 'MEMOBJ_STATEMENT')
                  begin
                        set @cstmt_count = @cstmt_count + 1
set @cstmt_size = @cstmt_size + (@pages * @page_size)
                  end
                  else if (@type = 'MEMOBJ_CURSOREXEC')
                  begin
                        set @cursor_count = @cursor_count + 1
set @cursor_size = @cursor_size + (@pages * @page_size)
                  end
                  else if (@type = 'MEMOBJ_QUERYEXECCNTXTFORSE')
                  begin
                        set @query_count = @query_count + 1
set @query_size = @query_size + (@pages * @page_size)
                  end
                  else if (@type = 'MEMOBJ_PLANSKELETON')
                  begin
                        set @skeleton_count = @skeleton_count + 1
                  end
                  
fetch next from detail_cursor into @type, @pages, @page_size
            end
 
            close detail_cursor
            deallocate detail_cursor
            
insert into PlanSummary values(@plan_handle, @current_plan_address, @total_size, @cstmt_count, @cstmt_size, @xstmt_count, @xstmt_size, @cursor_count, @cursor_size, @mxc_count, @mxc_size, @query_count, @query_size, @skeleton_count)
                  
            drop table #plan_details
            fetch next from plan_cursor into @current_plan_address
      end
 
      close plan_cursor
      deallocate plan_cursor
end
go
